from collections import OrderedDict
from tqdm import tqdm
import torch
from torch.cuda.amp import autocast
from .progress import ProgressBar
from .checkpoint import Monitor
import os
from pathlib import Path
from time import time
import torch.distributed as dist
from torch.distributed.fsdp import FullStateDictConfig, StateDictType
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP


class Trainer():
    '''
    self.lr
    self._epoch
    self.step
    '''

    def __init__(self,
                 task,
                 train_num_steps,
                 device,
                 validation_step,
                 save_ckpt,
                 monitor,
                 ckpt = None,
                 logger = None,
                 inference = False):


        self.task = task
        self.train_num_steps = train_num_steps
        self.ckpt = Monitor(**monitor)
        self.save_nums = monitor['save_top_k']
        self.validation_step = validation_step
        self.save_ckpt = save_ckpt
        if logger is not None:
            self.logger = logger
            self.log_id = logger.dir.split('/')[-2].split('-')[-1]

        else:
            self.logger = None
            self.log_id = 'off_'+str(int(time()//10))
            self.ckpt.ckpt_path = os.path.join('./checkpoint', self.log_id)

        if not os.path.exists('./checkpoint'):
            os.makedirs('./checkpoint')
        results_folder = os.path.join('./checkpoint', self.log_id)
        self.results_folder = Path(results_folder)
        self.results_folder.mkdir(exist_ok = True)
        self.task.results_folder = self.results_folder

        self.gradient_accumulate_every = 1
        self.gpu_availables = [x for x in range(torch.cuda.device_count())]

        if task.args.inference:
            for module in task.children():
                module = module.to(device)
                self.device_for_inf = device
        else:
            for module in task.children():
                module = module.to(dist.get_rank())

        if not task.args.inference:
            self.trainLoader = task.train_dataloader()
            self.valLoader = task.val_dataloader()
            self.evalLoader = task.test_dataloader()

        self.step = 0
        self.train_outputs = []
        self.val_outputs = []
        self.eval_outputs = []
        self.log_history = {}

        # task.load_from_state_dicts(torch.load('/home/gook/Local/E2ETTS/diffe2e_dis/checkpoint/anneal_dis/task-390000.pt', map_location = 'cpu'))
    def run(self):
        if dist.get_rank() == 0:
            self.summary()
        self._progress = ProgressBar(self.task, self.train_num_steps, initial = self.step)
        if self.task.automatic_optimization:
            self.task.optimizer.zero_grad()
        while self.step < self.train_num_steps:
            self.task.on_training_start()
            self.log(self.task.logs)
            self.reset_log('step')
            self._progress.on_training_start()
            self.model_to_train()

            batch = next(self.trainLoader)
            batch = [x.to(dist.get_rank(), non_blocking=True) if type(x) == torch.Tensor else x for x in batch]
            self.task.before_training_step()
            # forward
            output = self.task.training_step(batch)
            self.train_outputs.append(output)
            # backward
            if self.task.automatic_optimization:
                self.task.amp_scaler.scale(output['loss'] / self.gradient_accumulate_every).backward()
                if self.step%self.gradient_accumulate_every == 0:
                    self.task.amp_scaler.unscale_(self.task.optimizer)
                    torch.nn.utils.clip_grad_norm_(self.task.parameters(), 5)
                    self.task.amp_scaler.step(self.task.optimizer)
                    self.task.amp_scaler.update()
                    self.task.optimizer.zero_grad()
            self._lr_update(self.task.scheduler)
            self._progress.on_training_end(output['loss'])

            self.task.training_step_end(self.train_outputs)
            self.log(self.task.logs)
            self.reset_log('epoch')

            if dist.get_rank() == 0:
                if self.step != 0 and self.step % self.save_ckpt == 0:
                    self.save(milestone=self.step, save_nums = self.save_nums)

                if self.step != 0 and self.step % self.validation_step == 0:
                    self.model_to_eval()
                    with torch.no_grad():
                        self._progress.on_validation_start()
                        for batch_idx, batch in enumerate(self.valLoader):

                            batch = [x.to(dist.get_rank()) if type(x) == torch.Tensor else x for x in batch]
                            output = self.task.validation_step(batch)
                            self.val_outputs.append(output)

                            self._progress.on_validation_batch_end(output)
                            self.log(self.task.logs)
                            self.reset_log('step')
                        self._progress.val_bar_reset()
                        self.task.validation_loop_end(self.val_outputs)
                    self.log(self.task.logs)
                    if self.ckpt is not None:
                        self.ckpt.monitoring(self.task, self.step)
                    self.reset_log('epoch')


#==========================================================

    def _lr_update(self, scheduler):
        self.step += 1
        self.task.milestone = self.step
        if self.task.automatic_scheduler:
            self.lr = self.get_lr(self.task.optimizer)
        else:
            self.lr = self.task.lr

        if scheduler is not None:
            scheduler.step()

    def get_lr(self, optimizer):
        for param_group in optimizer.param_groups:
            return param_group['lr']

    def reset_log(self, level):
        if level == 'epoch':
            self.task.logs = {}
            self.train_outputs = []
            self.val_outputs = []
            self.eval_outputs = []

        elif level == 'step':
            self.task.logs = {}

    def model_to_train(self):
        for module in self.task.children():
            module = module.train()

    def model_to_eval(self):
        for module in self.task.children():
            module = module.eval()

    def save(self, milestone, save_nums = 3):
        data = self.task.state_dicts_list(milestone)

        if dist.get_rank() == 0:
            if os.path.isfile(str(self.results_folder / f'task-{milestone-self.save_ckpt*save_nums}.pt')):
                os.remove(str(self.results_folder / f'task-{milestone-self.save_ckpt*save_nums}.pt'))
        if self.task.args.fsdp:
            save_policy = FullStateDictConfig(offload_to_cpu=True, rank0_only=True)
            with FSDP.state_dict_type(self.task, StateDictType.FULL_STATE_DICT, save_policy):
                model_states = self.task.state_dicts_list(milestone)
            if dist.get_rank() == 0:
                torch.save(model_states, str(self.results_folder / f'task-{milestone}.pt'))
        else:
            if dist.get_rank() == 0:
                model_states = self.task.state_dicts_list(milestone)
                torch.save(model_states, str(self.results_folder / f'task-{milestone}.pt'))

    def load(self, log_id, milestone):
        results_folder = os.path.join('./checkpoint', log_id)
        results_folder = Path(results_folder)

        model_states = torch.load(str(results_folder / f'task-{milestone}.pt'), map_location = 'cuda:{}'.format(dist.get_rank()))
        # self.task.load_from_state_dicts(torch.load('/home/gook/Local/E2ETTS/diffe2e_dis/checkpoint/anneal_dis/task-390000.pt', map_location = 'cpu'))
        # model_states = torch.load(str(results_folder / f'task-{milestone}.pt'), map_location = 'cuda:{}'.format(self.device_for_inf))
        self.task.load_from_state_dicts(model_states)
        self.step = model_states['step']
        self._progress = ProgressBar(self.task, self.train_num_steps, initial = self.step)

    # ================ model summary ===============
    def summary(self):
        self._log_device_info(dist.get_rank())
        self._model_info()

    def _log_device_info(self, device):
        if device in ['cpu', 'CPU']:
            pass
        elif type(device) != int:
            raise Warning('type of device should be integer for cude:{}')

        _device_type = 'gpu' if device not in ['cpu', 'CPU'] else 'cpu'
        # _status = 'cpu' if device in ['cpu', 'CPU'] else torch.cuda.current_device()
        _status = 'cpu' if device in ['cpu', 'CPU'] else dist.get_rank()

        print(f"\nGPU available: {torch.cuda.is_available()}, used: {_device_type == 'gpu'}")
        print('LOCAL_RANK: {} - CUDA_VISIBLE_DEVICES: {}'.format(_status, [x for x in range(torch.cuda.device_count())]))
        if device not in ['cpu', 'CPU']:
            print('Device information : {}\n'.format(torch.cuda.get_device_name(device)))

    def _model_info(self):
        summary = OrderedDict((name, module) for name, module in list(self.task.named_children()))
        summary_value = [layer.__class__.__name__ for layer in summary.values()]

        len_name = max(len(layer) for layer in summary.keys())
        len_mod = max(len(layer.__class__.__name__) for layer in summary.values())

        col_widths = [2, len_name+3, len_mod+3, 8]
        cols = [[' ', []], ['Name', []], ['Type', []], ['#Params', []]]
        total_width = sum(col_widths) + 3*len(col_widths)
        n_rows = len(summary)

        for idx, key in enumerate(summary):
            name = key
            type = summary[key].__class__.__name__
            params = self.get_size(summary[key])

            cols[0][1].append(idx)
            cols[1][1].append(name)
            cols[2][1].append(type)
            cols[3][1].append(params)

        s = "{:<{}}"
        header = [s.format(c[0], l) for c, l in zip(cols, col_widths)]
        sums = " | ".join(header) + "\n" + "-" * total_width
        for i in range(n_rows):
            line = []
            for c, l in zip(cols, col_widths):
                line.append(s.format(str(c[1][i]), l))
            sums += "\n" + " | ".join(line)
        sums += "\n" + "-" * total_width
        print(sums)


    def get_size(self, model):
        pp=0
        for p in list(model.parameters()):
            nn=1
            for s in list(p.size()):
                nn = nn*s
            pp += nn

        return self.byteup(pp)

    def byteup(self, fsize):
        units = ['', ' K', ' M', ' T']
        for idx in range(0, 4):
            if fsize >= 1000:
                fsize /= 1000
            else:
                break
        return str(round(fsize, 2))+units[idx]


# ================ Logger ===============
    def log(self, data):
        if data:
            if self.logger is not None:
                data['step'] = self.step
                self.logger.log(data)

    def save_log(self, data):
        for key in data.keys():
            if key not in self.log_history.keys():
                self.log_history[key] = []
            self.log_history[key].append(data[key])



import os
import shutil

def file_backup(
            prtpath = '.',
            run_file_dir = 'run_code',
            ckpt_dir_name = 'ckpt'
            ):
    filelist = os.listdir(prtpath)
    for dir in ['wandb', 'checkpoint']:
        if dir in filelist:
            filelist.remove(dir)
    code_path = run_file_dir
    os.mkdir(code_path)
    print('>>> backup dir is made in \n{}'.format(run_file_dir))

    for file in filelist:
        path = os.path.join(prtpath, file)
        if os.path.isdir(file):
            shutil.copytree(path, os.path.join(code_path, file))
        elif os.path.isfile(file):
            shutil.copy(path, os.path.join(code_path, file))
    print('>>> run files are saved in \n{}'.format(run_file_dir))

    ckpt_path = os.path.join(code_path, ckpt_dir_name)
    os.mkdir(ckpt_path)
    print('>>> ckpt dir is made in \n{}'.format(ckpt_path))
    return run_file_dir, ckpt_path



import torch
from tqdm.auto import tqdm

class ProgressBar():
    def __init__(self, task, epoch, initial):
        self.epoch_len = epoch
        # self.train_step_len = task.train_dataloader().__len__()
        self.valid_step_len = task.val_dataloader().__len__()
        if task.test_dataloader():
            self.eval_step_len = task.test_dataloader().__len__()
        else: self.eval_step_len = 0

        # self.total_step_len = self.train_step_len + self.valid_step_len
        self.cur_step = initial
        self.loss = 0

        self.Training_bar = tqdm(
                    initial = self.cur_step,
                    leave = False,
                    desc = 'Training steps {}'.format(self.cur_step),
                    total = self.epoch_len,
                    postfix = {'loss': self.loss},
                    dynamic_ncols = True
                         )
        self.Validation_bar = tqdm(
                    leave = False,
                    desc = 'On validating',
                    total = self.valid_step_len,
                    dynamic_ncols = True,
                    postfix = {'loss': self.loss}
                         )
        # self.Evaluation_bar = tqdm(
        #             leave = False,
        #             desc = 'Evaluating',
        #             total = self.eval_step_len,
        #             dynamic_ncols = True,
        #                  )

    def on_training_start(self):
        self.cur_step += 1
        self.Training_bar.display()

    def on_training_end(self, output):
        if type(output) in [list, tuple]:
            self.loss = [round(op.detach().item(), 4) for op in output]
        else:
            self.loss = output.detach().item()
        self.Training_bar.update(1)
        self.Training_bar.set_description('Training step {}'.format(self.cur_step))
        self.Training_bar.set_postfix({'loss': self.loss})

    def on_train_batch_end(self, output):
        self.loss = output.detach().item()



    def on_validation_start(self):

        self.Validation_bar.set_description('Validation step {}'.format(self.cur_step))
        self.Validation_bar.display()

    def on_validation_batch_end(self, output):
        self.loss = 0
        # self.loss = output['loss'].detach().item()
        self.Validation_bar.update(1)
        self.Validation_bar.set_postfix({'loss': self.loss})

    def val_bar_reset(self):
        self.Validation_bar.refresh()
        self.Validation_bar.reset()



from collections import OrderedDict
from tqdm import tqdm
import torch
import torch.nn as nn
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
from torch.distributed.fsdp import FullStateDictConfig, StateDictType

class ExperimentSetup(nn.Module):
    '''
    self.model : Checkpoint로 저장되는 인스턴스



    (변경중)
    self.on_save_checkpoint: checkpoint를 저장할 모델을 return (default: self.model) -> dict
        key값은 저장되는 파일명을 결정
    '''


    def __init__(self, save_dir = None):
        self.optimizer = None
        super().__init__()
        self.model = None
        self.model_dict = None
        self.loss_func = None
        self.logs = {}
        self.save_dir = save_dir
        self.milestone = 0
        self.results_folder = None
        self.automatic_optimization = True
        self.automatic_scheduler = True
        self.optimizer = None
        self.scheduler = None
        self.amp_scaler = None
        
        if not self.automatic_scheduler:
            print('''
                  ===========***===========
                  You must define your own Class methods:
                  1. def get_lr(self):
                  2. def state_dicts_list(self, step):
                  3. def load_from_state_dicts(self, data):
                  ===========***===========
                  ''')
        # if self.model_dict is None:
        #     raise Warning("You should define self.model_dict = {'model': self.model} for save checkpoint")

    # def count_param(self):
    #     return func(self.model)

    def cycle(self, dl):
        while True:
            for data in dl:
                yield data

    def train_dataloader(self):
        return
    def val_dataloader(self):
        return
    def test_dataloader(self):
        return

    def augmentation(self, batch):
        return batch
    def configure_optimizers(self):
        '''
        self.optimizer =

        self.scheduler = {
                        'scheduler': ,
                        'interval': 'epoch' or 'step'
                        }
                        or None

        return self.optimizer, self.scheduler

        '''
        return

    def on_training_start(self):
        '''
        log all parameters what you want when start epoch
        '''
        self.log({'learning rate': self.get_lr()})
        self.lr = self.get_lr()


    def training_step(self, batch):
        # return {'loss': loss}
        return

    def before_training_step(self):
        return None

    def training_step_end(self, outputs):
        # torch.stack([x['loss'] for x in outputs]).mean()
        return None

    def validation_step(self, batch):
        # return {'loss': loss}
        return
    def validation_loop_end(self, outputs):
        # torch.stack([x['loss'] for x in outputs]).mean()
        return None

    def evaluation_step(self, batch):
        return None

    def evaluation_epoch_end(self):
        return None

    def evaluation_step_per_item(self, batch):
        '''
        batch의 각 요소의 shape에 batch size는 들어있지 않음
        즉, batch는 dataset의 get item으로 받아옴 (batch size = 1이고 squeeze된 상태)
        '''
        return None

    def evaluation_epoch_per_item_end(self):
        return None


    def state_dicts_list(self, step):
        if self.args.fsdp:
            save_policy = FullStateDictConfig(offload_to_cpu=True, rank0_only=True)
            with FSDP.state_dict_type(self.ddp_model, StateDictType.FULL_STATE_DICT, save_policy):
                task_state = self.ddp_model.state_dict()
                amp_state = self.amp_scaler.state_dict()
                if self.scheduler is not None:
                    sch_state = self.scheduler.state_dict()
        else:
            task_state = self.state_dict()
            amp_state = self.amp_scaler.state_dict()
            if self.scheduler is not None:
                sch_state = self.scheduler.state_dict()
        
        if self.scheduler is not None:
            data = {
                'step': step,
                'task': task_state,
                'scaler': amp_state,
                'lr_scheduler': sch_state
            }
        else:
            data = {
                'step': step,
                'task': task_state,
                'scaler': amp_state
            }

        return data

    def load_from_state_dicts(self, data):
        self.load_state_dict(data['task'], strict = False)
        self.amp_scaler.load_state_dict(data['scaler'])
        if self.scheduler is not None:
            self.scheduler.load_state_dict(data['lr_scheduler'])

    def load_from_checkpoint(self, device, ckpt_dict):
        '''
        ckpt_dict : {'모델의 변수명': 모델의 checkpoint path}
        '''

        for key in ckpt_dict.keys():
            for name, child in self.named_children():
                if key == name:
                    child.load_state_dict(torch.load(ckpt_dict[key], map_location = device))



    # ===============================

    def log(self, data):
        for key in data.keys():
            self.logs[key] = data[key]


    # ===============================

    def get_lr(self):
        if self.scheduler is not None:
            return self.scheduler.get_lr()[0]
        else:
            return self.optimizer.param_groups[-1]['lr']

    # ===============================

    def debugging(self, device):
        print('\n>>> Model debugging...')
        with torch.no_grad():
            tr_loader = self.train_dataloader()
            val_loader = self.val_dataloader()
            trainList = []
            valList = []
            print('.')
            for module in self.children():
                module = module.to(device)
            for batch_idx, batch in enumerate(tr_loader):
                batch = [x.to(device) if type(x) == torch.Tensor else x for x in batch]
                trainList.append(self.training_step(batch))
                break
            self.training_epoch_end(trainList)
            print('.')

            for batch_idx, batch in enumerate(val_loader):
                batch = [x.to(device) if type(x) == torch.Tensor else x for x in batch]
                valList.append(self.validation_step(batch))
                break
            self.validation_epoch_end(valList)
            print('.')
        torch.cuda.empty_cache()
        print(' Complete.\n')



import torch
import os
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
from torch.distributed.fsdp import FullStateDictConfig, StateDictType

class Monitor():
    '''
    monitor: checkpoint를 저장하는 기준 -> str
        dict format의 self.log의 key값을 받음


    '''

    def __init__(self,
                save_top_k = 3,
                monitor = None,
                mode = 'max',
                ckpt = '.'):

        self.save_top_k = save_top_k
        self.monitor = monitor
        self.ckpt_path = ckpt
        self.mode = mode

        if self.mode in ['max', 'Max']:
            self.filter = min
        elif self.mode in ['min', 'Min']:
            self.filter = max

        if self.monitor is None:
            self.top_k_list = []
        else:
            self.top_k_list = {}
        self.count = 0


    def monitoring(self, Task, step):
        if self.monitor is None:
            self.top_k_list.append(step)
            if self.count < self.save_top_k:
                self.save_ckpt(Task, step)
            else:
                self.delete_ckpt(Task, self.top_k_list.pop(0))
                self.save_ckpt(Task, step)

        elif self.monitor not in Task.logs.keys():
            raise Warning('{} is not in Task log`s key'.format(self.monitor))
        else:
            if Task.logs[self.monitor] not in self.top_k_list.keys():
                if self.count < self.save_top_k:
                    self.top_k_list[Task.logs[self.monitor]] = [step]
                    self.save_ckpt(Task, step) #Task.model
                else:
                    if Task.logs[self.monitor] < self.filter(self.top_k_list):          # fileter가 바뀌면 부등호 방향도 바뀌어야하지 않나
                        pass
                    else:
                        # print('새로운 acc')
                        # print('delete step: {}, save step: {}'.format(self.top_k_list[self.filter(self.top_k_list)][0], step))
                        self.delete_ckpt(Task, self.top_k_list[self.filter(self.top_k_list)].pop(0))
                        if len(self.top_k_list[self.filter(self.top_k_list)]) == 0:
                            del self.top_k_list[self.filter(self.top_k_list)]
                        self.save_ckpt(Task, step)
                        self.top_k_list[Task.logs[self.monitor]] = [step]

            elif Task.logs[self.monitor] in self.top_k_list.keys():
                if self.count < self.save_top_k:
                    self.top_k_list[Task.logs[self.monitor]].append(step)
                    self.save_ckpt(Task, step) #Task.model
                else:
                    if Task.logs[self.monitor] < self.filter(self.top_k_list):
                        pass
                    else:
                        del_step = self.top_k_list[self.filter(self.top_k_list)].pop(0)
                        # print('기존에 있는 acc')
                        # print('delete step: {}, save step: {}'.format(del_step, step))
                        self.delete_ckpt(Task, del_step)
                        self.save_ckpt(Task, step)
                        self.top_k_list[Task.logs[self.monitor]].append(step)

                        if len(self.top_k_list[self.filter(self.top_k_list)]) == 0:
                            del self.top_k_list[self.filter(self.top_k_list)]
                        # print(self.top_k_list)
            # print('count: ', self.count)
            # print(self.top_k_list)
            # print('min: ', min(self.top_k_list), '\n')

    def save_ckpt(self, Task, step, end = False):
        if end:
            for model in Task.model_dict.keys():
                path = os.path.join(self.ckpt_path, model+'_'+'end.ckpt')
                if Task.args.fsdp:
                    save_policy = FullStateDictConfig(offload_to_cpu=True, rank0_only=True)
                    with FSDP.state_dict_type(Task.model_dict[model], StateDictType.FULL_STATE_DICT, save_policy):
                        model_state = Task.model_dict[model].state_dict()
                else:
                    model_state = Task.model_dict[model].state_dict()
                torch.save(model_state, path)
        else:
            for model in Task.model_dict.keys():
                path = os.path.join(self.ckpt_path, model+'_'+str(step)+'.ckpt')
                if Task.args.fsdp:
                    save_policy = FullStateDictConfig(offload_to_cpu=True, rank0_only=True)
                    with FSDP.state_dict_type(Task.model_dict[model], StateDictType.FULL_STATE_DICT, save_policy):
                        model_state = Task.model_dict[model].state_dict()
                else:
                    model_state = Task.model_dict[model].state_dict()
                torch.save(model_state, path)
            self.count += 1


    def delete_ckpt(self, Task, step):
        for model in Task.model_dict.keys():
            path = os.path.join(self.ckpt_path, model+'_'+str(step)+'.ckpt')
            if os.path.exists(path):
                os.remove(path)
        self.count -= 1



from collections import OrderedDict
from tqdm import tqdm
import torch
from .progress import ProgressBar
from glob import glob


class Eval():
    '''
    self.lr
    self._epoch
    self.step
    '''

    def __init__(self,
                 task,
                 device,
                 ckpt_dict,
                 ):


        self.task = task
        self.ckpt_dict = ckpt_dict
        self.device = device

        self.evalLoader = task.test_dataloader()
        self.device_num = device
        if device in ['cpu', 'CPU']:
            self.device = 'cpu'
        elif device in [0, 1, 2, 3]:
            self.device = 'cuda:{}'.format(device)


        for module in task.children():
            module = module.to(self.device)

        self.task.load_from_checkpoint(
                                    device = self.device,
                                    ckpt_dict = ckpt_dict
                                    )

    def run(self):
        with torch.no_grad():
            for batch in tqdm(self.evalLoader):
                batch = [x.to(self.device) for x in batch]
                self.task.evaluation_step(batch)

            self.task.evaluation_epoch_end()

        # with torch.no_grad():
        #     for batch in tqdm(range(self.valLoader.dataset.__len__())):
        #         batch = [x.to(self.device) for x in batch]
        #         self.task.evaluation_step(batch)
        #
        #     self.task.evaluation_epoch_end()
